@@ -1,20 +1,17 @@
-#  File R/misc.utilities.R in package statnet.common, part of the
-#  Statnet suite of packages for network analysis, https://statnet.org .
+#  File R/misc.utilities.R in package statnet.common, part of the Statnet suite
+#  of packages for network analysis, https://statnet.org .
 #
-#  This software is distributed under the GPL-3 license.  It is free,
-#  open source, and has the attribution requirements (GPL Section 7) at
+#  This software is distributed under the GPL-3 license.  It is free, open
+#  source, and has the attribution requirements (GPL Section 7) at
 #  https://statnet.org/attribution .
 #
-#  Copyright 2007-2024 Statnet Commons
+#  Copyright 2007-2025 Statnet Commons
 ################################################################################
 #' reorder vector v into order determined by matching the names of its elements
 #' to a vector of names
 #' 
-#' A helper function to reorder vector \code{v} (if named) into order specified
-#' by matching its names to the argument \code{names}
-#' 
-#' does some checking of appropriateness of arguments, and reorders v by
-#' matching its names to character vector \code{names}
+#' This function is deprecated in favor of [match_names()] and will be
+#' removed in a future release.
 #' 
 #' @param v a vector (or list) with named elements, to be reorderd
 #' @param names a character vector of element names, corresponding to names of
@@ -22,7 +19,7 @@
 #' @param errname optional, name to be reported in any error messages. default
 #' to \code{deparse(substitute(v))}
 #' @return returns \code{v}, with elements reordered
-#' @note earlier versions of this function did not order as advertiased
+#' @note earlier versions of this function did not order as advertised
 #' @examples
 #' 
 #' test<-list(c=1,b=2,a=3)
@@ -31,10 +28,10 @@
 vector.namesmatch<-function(v,names,errname=NULL){
   if(is.null(errname)) errname <- deparse(substitute(v))
 
-  if (is.null(names(v))){
+  if(is.null(names(v))){
     if(length(v) == length(names)){
       names(v) <- names
-    }else stop('Length of "', errname, '" is ', length(v), " but should be ", length(names),".")
+    }else stop('Length of ', sQuote(errname), ' is ', length(v), " but should be ", length(names),".")
   }else{
     if(length(v) == length(names)
        && length(unique(names(v)))==length(v)
@@ -42,11 +39,151 @@ vector.namesmatch<-function(v,names,errname=NULL){
        && all(sort(names(v)) == sort(names))){
       namesmatch <- match(names, names(v))
       v <- v[namesmatch]
-    }else stop('Name mismatch in "', errname,'". Specify by position.')
+    }else stop('Name mismatch in ', sQuote(errname),'. Specify by position.')
   }
   v
 }
 
+#' Construct a named vector with semantics useful for parameter vectors
+#'
+#' This is a helper function that constructs a named vector with names
+#' in `names` with values taken from `v` and optionally `default`,
+#' performing various checks. It supersedes [vector.namesmatch()].
+#'
+#' If `v` is not named, it is required to be the same length as
+#' `names` and is simply given the corresponding names. If it is
+#' named, nonempty names are matched to the corresponding elements of
+#' `names`, with partial matching supported.
+#'
+#' Default values can be specified by the caller in `default` or by
+#' the end-user by adding an element with an empty (`""`) name in
+#' addition to the others. If given, the latter overrides the former.
+#'
+#' Duplicated names in `v` or `names` are resolved sequentially,
+#' though note the example below for caveat about partial matching.
+#'
+#' Zero-length `v` is handled as follows:
+#'
+#' * If length of `names` is empty, return `v` unchanged.
+#'
+#' * If it is not and `default` is not `NULL`, return the `default` vector.
+#'
+#' * Otherwise, raise an error.
+#'
+#' An informative error is raised under any of the following conditions:
+#'
+#' * `v` is not named but has length that differs from that of `names`.
+#'
+#' * More than one element of `v` has an empty name.
+#'
+#' * Not all elements of `names` are matched by an element of `v`, and
+#'   no default is specified.
+#'
+#' * Not all elements of `v` are used up for elements of `names`.
+#'
+#' * There is ambiguity that [pmatch()] cannot resolve.
+#'
+#' @note At this time, passing `partial=FALSE` will use a crude
+#'   sentinel to prevent partial matching, which in some, extremely
+#'   improbable, circumstances might not work.
+#'
+#' @param v a vector
+#' @param names a character vector of element names
+#' @param default value to be used for elements of `names` not found in `v`
+#' @param partial whether partial matching is allowed
+#' @param errname optional, name to be reported in any error messages;
+#'   defaults to `deparse1(substitute(v))`
+#' @return A named vector with names `names` (in that order). See
+#'   Details.
+#' @examples
+#'
+#' # Unnamed:
+#' test <- as.numeric(1:3)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'c', 'b')),
+#'   c(a = 1, c = 2, b = 3)
+#' ))
+#'
+#' # Named, reordered:
+#' test <- c(c = 1, b = 2, a = 3)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'c', 'b')),
+#'   c(a = 3, c = 1, b = 2)
+#' ))
+#'
+#' # Default value specified by default= assigned to a
+#' test <- c(c = 1, b = 2)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'c', 'b'), NA),
+#'   c(a = NA, c = 1, b = 2)
+#' ))
+#'
+#' # Default value specified in v assigned to a and b:
+#' test <- c(c = 1, 2)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'c', 'b')),
+#'   c(a = 2, c = 1, b = 2)
+#' ))
+#'
+#' # Partial matching
+#' test <- c(c = 1, 2)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'cab', 'b')),
+#'   c(a = 2, cab = 1, b = 2)
+#' ))
+#'
+#' # Multiple matching
+#' test <- c(c = 1, 2, c = 3)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'c', 'c')),
+#'   c(a = 2, c = 1, c = 3)
+#' ))
+#'
+#' # Partial + multiple matching caveat: exact match will match first.
+#' test <- c(c = 1, a = 2, ca = 3)
+#' stopifnot(identical(
+#'   match_names(test, c('a', 'ca', 'ca')),
+#'   c(a = 2, ca = 3, ca = 1)
+#' ))
+#'
+#' @importFrom stats na.omit setNames
+#' @export
+match_names <- function(v, names, default = NULL, partial = TRUE, errname = NULL) {
+  if(is.null(errname)) errname <- deparse1(substitute(v))
+
+  if(length(v) == 0) {
+    if(length(names) == 0) v
+    else if(!is.null(default)) setNames(rep_len(default, length(names)), names)
+    else stop(sQuote(errname), ' is ', sQuote(deparse1(v)), " but should have ", length(names)," element(s):\n", paste(strwrap(paste(sQuote(names), collapse = ", "), indent = 2, exdent = 2), collapse = "\n"), call. = FALSE)
+  }else if(is.null(names(v))){
+    if(length(v) == length(names)){
+      setNames(v, names)
+    }else stop('Length of ', sQuote(errname), ' is ', length(v), " but should be ", length(names),":\n", paste(strwrap(paste(sQuote(names), collapse = ", "), indent = 2, exdent = 2), collapse = "\n"), call. = FALSE)
+  }else{
+    blanks <- names(v) == ""
+    if(any(blanks)){
+      if(sum(blanks) > 1L) stop("Named vector ", sQuote(errname), " may have at most one unnamed element.", call. = FALSE)
+
+      default <- unname(v[blanks])
+      v <- v[!blanks]
+    }
+
+    # partial == FALSE -> add a sentinel string at the end of all strings to prevent partial matching.
+    namesmatch <- if(partial) pmatch(names(v), names)
+                  else pmatch(paste(names(v), "\n\xf5\xdc\n"), paste(names, "\n\xf5\xdc\n"))
+    used <- !is.na(namesmatch)
+    found <- unwhich(na.omit(namesmatch), length(names))
+
+    if(is.null(default) && !all(found)) stop("Named vector ", sQuote(errname), " is missing values for the following elements: ", paste.and(sQuote(names[!found])), ".", call. = FALSE)
+    if(!all(used)) stop("In named vector ", sQuote(errname), " unused or not uniquely matched elements: ", substr(s <- deparse1(v[!used]), 3, nchar(s)-1L), call. = FALSE)
+
+    numeric(length(names)) |>
+      replace(na.omit(namesmatch), v[used]) |>
+      replace(!found, default) |>
+      setNames(names)
+  }
+}
+
 #' "Compress" a data frame.
 #' 
 #' \code{compress_rows.data.frame} "compresses" a data frame, returning unique rows
@@ -331,28 +468,25 @@ EVL3 <- function(test, notnull, null = NULL){
 }
 
 
-#' Return the first argument passed (out of any number) that is not a
-#' \code{try-error} (result of \code{\link[base]{try}} encountering an error.
+#' Attempt a series of statements and return the first one that is not an error.
+#'
+#' `ERRVL()` expects the potentially erring statements to be wrapped
+#' in [try()]. In addition, all expressions after the first may
+#' contain a `.`, which is substituted with the `try-error` object
+#' returned by the previous expression.
 #' 
-#' This function is inspired by \code{\link{NVL}}, and simply returns the first
-#' argument that is not a \code{try-error}, raising an error if all arguments
-#' are \code{try-error}s.
+#' @note This family of functions behave similarly to the [NVL()] and the [EVL()] families.
 #' 
-#' 
-#' @param \dots Expressions to be tested; usually outputs of
-#'   \code{\link[base]{try}}.
-#' @return The first argument that is not a \code{try-error}. Stops
-#'   with an error if all are.
-#' @note This function uses lazy evaluation, so, for example `ERRVL(1,
-#'   stop("Error!"))` will never evaluate the [`stop`] call and will
-#'   not produce an error, whereas `ERRVL(try(solve(0)),
+#' @param \dots Expressions to be attempted; for `ERRVL()`, should be
+#'   wrapped in [try()].
+#' @return The first argument that is not an error. Stops with an
+#'   error if all are.
+#' @note These functions use lazy evaluation, so, for example
+#'   `ERRVL(1, stop("Error!"))` will never evaluate the [stop()] call
+#'   and will not produce an error, whereas `ERRVL2(solve(0),
 #'   stop("Error!"))` would.
 #'
-#' In addition, all expressions after the first may contain a `.`,
-#' which is substituted with the `try-error` object returned by the
-#' previous expression.
-#' 
-#' @seealso \code{\link[base]{try}}, \code{\link[base]{inherits}}
+#' @seealso [try()], [inherits()], [tryCatch()]
 #' @keywords utilities
 #' @examples
 #' 
@@ -364,11 +498,11 @@ EVL3 <- function(test, notnull, null = NULL){
 #' # Error:
 #' print(ERRVL(try(solve(0), silent=TRUE),
 #'             stop("Error!")))
-#' 
-#' # Error with an elaborate message:
-#' print(ERRVL(try(solve(0), silent=TRUE),
-#'             stop("Stopped with an error: ", .)))
 #' }
+#'
+#' # Capture and print the try-error object:
+#' ERRVL(try(solve(0), silent=TRUE),
+#'       print(paste0("Stopped with an error: ", .)))
 #' @export
 ERRVL <- function(...){
   x <- NULL
@@ -379,6 +513,53 @@ ERRVL <- function(...){
   stop("No non-error expressions passed.")
 }
 
+#' @rdname ERRVL
+#' @description `ERRVL2()` does *not* require the potentially erring
+#'   statements to be wrapped in [try()] and will, in fact, treat them
+#'   as non-erring; it does not perform dot substitution.
+#'
+#' @examples
+#'
+#' print(ERRVL2(1,2,3)) # 1
+#' print(ERRVL2(solve(0),2,3)) # 2
+#' print(ERRVL2(1, stop("Error!"))) # No error
+#'
+#' @export
+ERRVL2 <- function(...){
+  for(e in eval(substitute(alist(...)))) # Lazy evaluate. (See http://adv-r.had.co.nz/Computing-on-the-language.html .)
+    tryCatch(return(eval(e, parent.frame())),
+             error = function(err){})
+  stop("No non-error expressions passed.")
+}
+
+#' @rdname ERRVL
+#' @description `ERRVL3()` behaves as `ERRVL2()`, but it does perform
+#'   dot-substitution with the [`condition`] object.
+#'
+#' @examples
+#'
+#' \dontrun{
+#' # Error:
+#' ERRVL3(solve(0), stop("Error!"))
+#' }
+#'
+#' # Capture and print the error object:
+#' ERRVL3(solve(0), print(paste0("Stopped with an error: ", .)))
+#'
+#' # Shorthand for tryCatch(expr, error = function(e) e):
+#' ERRVL3(solve(0), .)
+#'
+#' @export
+ERRVL3 <- function(...){
+  x <- NULL
+  for(e in eval(substitute(alist(...)))) # Lazy evaluate. (See http://adv-r.had.co.nz/Computing-on-the-language.html .)
+    x <- tryCatch(
+      return(eval(if(!is.null(x)) do.call(substitute, list(e, list(.=x)))
+                  else e,
+                  parent.frame())),
+      error = function(err) err)
+  stop("No non-error expressions passed.")
+}
 
 #' Optionally test code depending on environment variable.
 #' 
@@ -413,24 +594,51 @@ opttest <- function(expr, testname=NULL, testvar="ENABLE_statnet_TESTS", yesvals
 #' Test if all items in a vector or a list are identical.
 #'
 #' @param x a vector or a list
+#' @param .p a predicate function of two arguments returning a logical.
+#'   Defaults to [identical()].
+#' @param .ref integer; index of element of `x` to which all the remaining
+#'   ones will be compared. Defaults to 1.
+#' @param ... additional arguments passed to `.p()`
 #'
-#' @return `TRUE` if all elements of `x` are identical to each other.
+#' @return By default `TRUE` if all elements of `x` are identical to each
+#'   other, `FALSE` otherwise. In the general case, `all_identical()`
+#'   returns `TRUE` if and only if `.p()` returns `TRUE` for all the pairs
+#'   involving the first element and the remaining elements.
 #'
-#' @seealso [`identical`]
+#' @seealso [identical()], [all.equal()]
 #'
 #' @examples
 #'
 #' stopifnot(!all_identical(1:3))
 #'
 #' stopifnot(all_identical(list("a", "a", "a")))
+#' 
+#' # Using with `all.equal()` has its quirks 
+#' # because of numerical tolerance:
+#' x <- seq(
+#'   .Machine$double.eps, 
+#'   .Machine$double.eps + 1.1 * sqrt(.Machine$double.eps), 
+#'   length = 3
+#' )
+#' # Results with `all.equal()` are affected by ordering
+#' all_identical(x, all.equal) # FALSE
+#' all_identical(x[c(2,3,1)], all.equal) # TRUE
+#' # ... because `all.equal()` is intransitive
+#' all_identical(x[-3], all.equal) # is TRUE and
+#' all_identical(x[-1], all.equal) # is TRUE, but
+#' all_identical(x[-2], all.equal) # is FALSE
+#' 
 #' @export
-all_identical <- function(x){
-  if(length(x)==0) return(TRUE)
-  v0 <- x[[1L]]
-  for(v in x[-1]) if(!identical(v0,v)) return(FALSE)
+all_identical <- function(x, .p = identical, .ref = 1L, ...){
+  if(length(x) == 0) return(TRUE)
+  stopifnot(is.function(.p))
+  stopifnot(length(.ref) == 1)
+  v0 <- x[[.ref]]
+  for(v in x[- .ref]) if(!isTRUE(.p(v0, v, ...))) return(FALSE)
   return(TRUE)
 }
 
+
 #' Construct a logical vector with `TRUE` in specified positions.
 #'
 #' This function is basically an inverse of [`which`].
@@ -926,3 +1134,134 @@ unused_dots_warning <- function(e){
   rlang::warn(sprintf("Argument(s) %s were not recognized or used. Did you mistype an argument name?",
                paste.and(sQuote(v))))
 }
+
+#' Modify the argument in the calling environment of the calling function
+#'
+#' This is a helper function that enables a function to modify its argument in place, emulating behavior of \CRANpkg{R6} classes and methods in the \CRANpkg{network}. It should typically be the last line of the calling function.
+#'
+#' This function determines whether the argument can be assigned to by actually attempting to do so. If this results in an error, for example, because the argument is anonymous, the error is silently ignored.
+#'
+#' It can be called multiple times by the same function to modify multiple arguments. It uses the [on.exit()] mechanism, adding to the list. Thus, if some other function calls `on.exit(..., add = FALSE)` (the default) afterwards, `modify_in_place()` will fail silently.
+#'
+#' @param x the argument (not its name!) to be modified
+#' @param value the value to assign (defaulting to the current value of `x`)
+#'
+#' @return `value`, invisibly, while attempting to modify `x` in place
+#'
+#' @examples
+#' ## A function that increments its argument in place:
+#' inc <- function(x){
+#'   modify_in_place(x, x+1)
+#' }
+#'
+#' y <- 1
+#' z <- 1
+#'
+#' stopifnot(inc(z) == 2)
+#' stopifnot(z == 2)
+#' stopifnot(inc(y) == 2)
+#' stopifnot(y == 2)
+#' stopifnot(inc(z) == 3)
+#' stopifnot(z == 3)
+#'
+#' stopifnot(inc(identity(z)) == 4)
+#' stopifnot(z == 3) # Not updated!
+#'
+#' ## Modify an argument that's been updated in place:
+#' inc2 <- function(y){
+#'   y <- y + 1
+#'   modify_in_place(y)
+#' }
+#'
+#' z
+#' stopifnot(inc2(z) == 4)
+#' stopifnot(z == 4)
+#'
+#' ## Decrement the first argument, increment the second:
+#' incdec <- function(x,y){
+#'   modify_in_place(x, x-1)
+#'   modify_in_place(y, y+1)
+#' }
+#'
+#' c(y,z)
+#' incdec(y,z)
+#' stopifnot(all(c(y,z) == c(1,5)))
+#' @export
+modify_in_place <- function(x, value = x){
+  xn <- substitute(x) # Grab the name of the argument to be updated.
+  xnn <- match.call(sys.function(-1), sys.call(-1))[[xn]] # Grab the expression that was passed into its argument.
+
+  eval.parent(on.exit( # As the calling function exits...
+    tryCatch( # try to...
+    eval.parent(call("<-", xnn, value), n = 2), # Assign `value` to whatever `xnn` stands for in the caller's calling environment.
+    error = identity # and do nothing if it fails.
+    ), add = TRUE))
+
+  invisible(value) # Return invisibly.
+}
+
+#' Replace values in a vector according to functions
+#'
+#' This is a thin wrapper around [base::replace()] that allows `list`
+#' and/or `values` to be functions that are evaluated on `x` to obtain
+#' the replacement indices and values. The assignment version replaces
+#' `x`.
+#'
+#' `list` function is passed the whole vector `x` at once (not
+#' elementwise) and any additional arguments to `replace()`, and must
+#' return an indexing vector (numeric, logical, character,
+#' etc.). `values`/`value` function is passed `x` after subsetting it by the
+#' result of calling `list()`.
+#'
+#' If passing named arguments, `x`, `list`, and `values` may cause a
+#' conflict.
+#'
+#' @param x a vector.
+#' @param list either an index vector or a function (*not* a function
+#'   name).
+#' @param values,value either a vector of replacement values or a function
+#'   (*not* a function name).
+#' @param ... additional arguments to `list` if it is a function;
+#'   otherwise ignored.
+#'
+#' @return A vector with the values replaced.
+#'
+#' @seealso [purrr::modify()] family of functions.
+#'
+#' @examples
+#'
+#' (x <- rnorm(10))
+#'
+#' ### Replace elements of x that are < 1/4 with 0.
+#'
+#' # Note that this code is pipeable.
+#' x |> replace(`<`, 0, 1/4)
+#' # More readable, using lambda notation.
+#' x |> replace(\(.x) .x < 1/4, 0)
+#' # base equivalent.
+#' stopifnot(identical(replace(x, `<`, 0, 1/4),
+#'                     base::replace(x, x < 1/4, 0)))
+#'
+#' ### Multiply negative elements of x by 1i.
+#'
+#' x |> replace(\(.x) .x < 0, \(.x) .x * 1i)
+#' stopifnot(identical(replace(x, \(.x) .x < 0, \(.x) .x * 1i),
+#'                     base::replace(x, x < 0, x[x < 0] * 1i)))
+#'
+#' ### Modify the list in place.
+#'
+#' y <- x
+#' replace(x, `<`, 1/4) <- 0
+#' x
+#' stopifnot(identical(x, replace(y, `<`, 0, 1/4)))
+#'
+#' @export
+replace <- function(x, list, values, ...) {
+  if (is.function(list)) list <- list(x, ...)
+  if (is.function(values)) values <- values(x[list], ...)
+  base::replace(x, list, values)
+}
+
+#' @rdname replace
+#' @export
+`replace<-` <- function(x, list, ..., value) replace(x, list, value, ...)
